--imports
local tablex = require("pl.tablex")
local StmtType = require("dogma.syn.StmtType")
local transName = require("dogma.trans.js._.util").transName
local SubTrans = require("dogma.trans.js._.SubTrans")
local SentType = require("dogma.syn.SentType")
local DirectiveType = require("dogma.syn.DirectiveType")
local NodeType = require("dogma.syn.NodeType")
local NonTerminalType = require("dogma.syn.NonTerminalType")
local TerminalType = require("dogma.syn.TerminalType")

--A statement transformer.
local StmtTrans = {}
StmtTrans.__index = StmtTrans
setmetatable(StmtTrans, {__index = SubTrans})
package.loaded[...] = StmtTrans

--Constructor.
--
--@param trans:Trans  Parent transformer.
function StmtTrans.new(trans)
  return setmetatable(SubTrans.new(trans), StmtTrans)
end

--Transform a statement.
--
--@param stmt:Stmt  Statement to transform.
--@return string
function StmtTrans:transform(stmt)
  if stmt.subtype == StmtType.ASYNC then
    return self:_transAsync(stmt)
  elseif stmt.subtype == StmtType.BREAK then
    return self:_transBreak(stmt)
  elseif stmt.subtype == StmtType.DO then
    return self:_transDo(stmt)
  elseif stmt.subtype == StmtType.ENUM then
    return self:_transEnum(stmt)
  elseif stmt.subtype == StmtType.EXPORT then
    return self:_transExport(stmt)
  elseif stmt.subtype == StmtType.FN then
    return self:_transFn(stmt)
  elseif stmt.subtype == StmtType.FOR then
    return self:_transFor(stmt)
  elseif stmt.subtype == StmtType.FOR_EACH then
    return self:_transForEach(stmt)
  elseif stmt.subtype == StmtType.FROM then
    return self:_transFrom(stmt)
  elseif stmt.subtype == StmtType.IF then
    return self:_transIf(stmt)
  elseif stmt.subtype == StmtType.NEXT then
    return self:_transNext(stmt)
  elseif stmt.subtype == StmtType.PROC then
    return self:_transProc(stmt)
  elseif stmt.subtype == StmtType.PUB then
    return self:_transPub(stmt)
  elseif stmt.subtype == StmtType.RETURN then
    return self:_transReturn(stmt)
  elseif stmt.subtype == StmtType.TYPE then
    return self:_transType(stmt)
  elseif stmt.subtype == StmtType.STRUCT then
    return self:_transStruct(stmt)
  elseif stmt.subtype == StmtType.USE then
    return self:_transUse(stmt)
  elseif stmt.subtype == StmtType.VAR then
    return self:_transVar(stmt)
  elseif stmt.subtype == StmtType.WHILE then
    return self:_transWhile(stmt)
  elseif stmt.subtype == StmtType.WITH then
    return self:_transWith(stmt)
  elseif stmt.subtype == StmtType.YIELD then
    return self:_transYield(stmt)
  end
end

--Transform a use statement.
--
--@return string
function StmtTrans:_transUse(stmt)
  local code

  --(1) transform
  code = ""

  for _, mod in ipairs(stmt.modules) do
    code = code .. string.format('const %s = dogma.use(require("%s"));', mod.name, mod.path, mod.path)

    if mod.uVars then
      code = code .. "let {"

      for i, name in ipairs(mod.uVars) do
        code = code .. (i == 1 and "" or ",") .. name
      end

      code = code .. string.format("} = %s;", mod.name)
    end
  end

  --(2) return
  return code
end

--Transform a from statement.
--
--@return string
function StmtTrans:_transFrom(stmt)
  local code

  --(1) transform
  code = "import {"

  for i, mem in ipairs(stmt.members) do
    code = code .. (i == 1 and "" or ", ")

    if mem.name == mem.as then
      code = code .. mem.name
    else
      code = code .. string.format("%s as %s", mem.name, mem.as)
    end
  end

  code = code .. '} from "' .. stmt.module .. '";'

  for _, mem in ipairs(stmt.members) do
    if mem.uVars then
      code = code .. "let {"

      for i, var in ipairs(mem.uVars) do
        code = code .. (i == 1 and "" or ",") .. var
      end

      code = code .. string.format("} = %s;", mem.as)
    end
  end

  --(2) return
  return code
end

--Transform a break statement.
--
--@return string
function StmtTrans:_transBreak(stmt)
  return "break;"
end

--Transform a next statement.
--
--@return string
function StmtTrans:_transNext(stmt)
  return "continue;"
end

--Transform an enum statement.
--
--@return string
function StmtTrans:_transEnum(enum)
  local code

  --(1) transform
  code = string.format([[
class %s {
  constructor(name, val) {
    Object.defineProperty(this, "name", {value: name, enum: true});
    Object.defineProperty(this, "value", {value: val, enum: true});
  }

  /* istanbul ignore next */
  toString() {
    return this.name;
  }
}
]],
    enum.name
  )

  code = code .. self:_transVisib(enum.visib, enum.name)

  for _, item in ipairs(enum.items) do
    code = code .. string.format(
      [[Object.defineProperty(%s, "%s", {value: new %s("%s", %s), enum: true});]] .. "\n",
      enum.name,
      item.name,
      enum.name,
      item.name,
      (type(item.value) == "string" and string.format([["%s"]], item.value) or item.value)
    )
  end

  --(2) return
  return code
end

function StmtTrans:_transVisib(visib, item)
  if visib == "export" then
    return "module.exports = exports = " .. item .. ";"
  elseif visib == "pub" then
    if type(item) == "table" then
      item = table.concat(item, ", ")
    end

    return "export {" .. item .. "};"
  else
    return ""
  end
end

--Transform an async statement.
--
--@param node
--@return string
function StmtTrans:_transAsync(node)
  local trans = self._.trans
  local code

  --(1) get sentences to run
  if node.catch then
    code = "{try "
  else
    code = ""
  end

  code = code .. self:_transBody(node.body)

  if node.catch then
    code = code .. self:_transCatch(node.catch)
    code = code .. "}"
  end

  --(2) return
  if node.opts["repeat"] then
    return string.format(
      "dogma.repeatAsyncly(async () => %s, {delay: %s, repeat: %s});",
      code,
      node.opts.delay and trans:_trans(node.opts.delay) or "null",
      trans:_trans(node.opts["repeat"])
    )
  elseif node.opts.delay then
    return string.format("dogma.asyncly(async () => %s, %s);", code, trans:_trans(node.opts.delay))
  else
    return string.format("setImmediate(async () => %s);", code)
  end
end

--Transform a var statement.
--
--@return string
function StmtTrans:_transVar(stmt)
  local code

  --(1) transform
  code = ""

  if #stmt.decls > 0 then
    for _, decl in ipairs(stmt.decls) do
      if decl.subtype == "std" then
        code = code .. self:_transStdDecl(stmt.visib, decl)
      elseif decl.subtype == "map" then
        code = code .. self:_transMapDecl(stmt.visib, decl)
      else
        code = code .. self:_transListDecl(stmt.visib, decl)
      end
    end
  end

  --(2) return
  return code
end

function StmtTrans:_transStdDecl(visib, decl)
  local code

  --(1) transform
  code = (decl.type == "var" and (visib and "var " or "let ") or "const ")

  if decl.value then
    code = code .. string.format("%s = %s;", decl.names[1], self._.trans:_trans(decl.value))
  else
    code = code .. decl.names[1] .. ";"
  end

  code = code .. self:_transVisib(visib, decl.names[1])

  --(2) return
  return code
end

function StmtTrans:_transMapDecl(visib, decl)
  local code

  --(1) transform
  code = (decl.type == "var" and (visib and "var {" or "let {") or "const {")

  for ix, name in ipairs(decl.names) do
    code = code .. (ix == 1 and "" or ", ") .. name
  end

  code = code .. string.format("} = %s;", self._.trans:_trans(decl.value))

  if visib then
    code = code .. self:_transVisib(visib, decl.names)
  end

  --(2) return
  return code;
end

function StmtTrans:_transListDecl(visib, decl)
  local code

  --(1) transform
  code = (decl.type == "var" and (visib and "var [" or "let [") or "const [")

  for ix, name in ipairs(decl.names) do
    code = code .. (ix == 1 and "" or ", ") .. name
  end

  code = code .. string.format(
    "] = dogma.getArrayToUnpack(%s, %s);",
    self._.trans:_trans(decl.value),
    #decl.names
  )

  if visib then
    code = code .. self:_transVisib(visib, decl.names):gsub("%.%.%.", "")
  end

  --(2) return
  return code;
end

--Transform an while statement.
--
--@return string
function StmtTrans:_transWhile(node)
  local trans = self._.trans
  local code

  --(1) transform
  if node.iter then
    code = string.format(
      "for (; %s; %s) ",
      trans:_trans(node.cond),
      trans:_trans(node.iter)
    )
  else
    code = string.format("while (%s) ", trans:_trans(node.cond))
  end

  if node.catch or node.finally then
    code = code .. "try "
  end

  code = code .. self:_transBody(node.body)
  code = code .. self:_transCatch(node.catch)
  code = code .. self:_transFinally(node.finally)

  --(2) return
  return code
end

--Transform a statement body.
--
--2return string
function StmtTrans:_transBody(body)
  local trans = self._.trans
  local code

  --(1) transform
  local sent, covDir, i

  code = "{"

  i = 1
  while i <= #body do
    sent = body[i]

    if sent.type == SentType.DIRECTIVE and sent.subtype == DirectiveType.COVERAGE then
      covDir = sent
      i = i + 1
      sent = body[i]

      if sent == nil then
        sent = covDir
        covDir = nil
      end
    else
      covDir = nil
    end

    code = code .. trans:_trans(sent, ";", covDir)

    i = i + 1
  end

  code = code .. "}"

  --(2) return
  return code
end

--Transform a catch clause.
--
--@return string
function StmtTrans:_transCatch(catch)
  local trans = self._.trans
  local code

  --(1) transform
  if catch then
    code = string.format(" catch(%s) {", catch.var or "_")

    for _, sent in ipairs(catch.body) do
      code = code .. trans:_trans(sent, ";")
    end

    code = code .. "}"
  end

  --(2) return
  return code or ""
end

--Transform a finally clause.
--
--@return string
function StmtTrans:_transFinally(fin)
  local trans = self._.trans
  local code

  --(1) transform
  if fin then
    code = " finally {"

    for _, sent in ipairs(fin.body) do
      code = code .. trans:_trans(sent, ";")
    end

    code = code .. "}"
  end

  --(2) return
  return code or ""
end

--Transform a do statement.
--
--@return string
function StmtTrans:_transDo(node)
  local trans = self._.trans
  local code

  --(1) transform
  if #node.vars == 0 then
    code = ""
  else
    code = string.format("{let %s;", table.concat(node.vars, ", "))
  end

  if node.catch or node.finally then
    if node.cond then
      code = code .. "do try "
    else
      code = code .. "try "
    end
  else
    code = code .. "do "
  end

  code = code .. self:_transBody(node.body)

  code = code .. self:_transCatch(node.catch)
  code = code .. self:_transFinally(node.finally)

  if node.cond then
    code = code .. string.format(" while (%s);", trans:_trans(node.cond))
  end

  if #node.vars > 0 then
    code = code .. "}"
  end

  --(2) return
  return code
end

--Transfor a for statement.
--
--@return string
function StmtTrans:_transFor(stmt)
  local trans = self._.trans
  local code

  --(1) transform
  code = "for (let "

  for i, var in ipairs(stmt.def) do
    code = code .. (i == 1 and "" or ", ") .. var.name

    if var.value then
      code = code .. " = " .. trans:_trans(var.value)
    end
  end

  code = code .. "; " .. trans:_trans(stmt.cond) .. "; "

  if stmt.iter then
    code = code .. trans:_trans(stmt.iter)
  end

  code = code .. ") "

  if stmt.catch or stmt.finally then
    code = code .. "try "
  end

  code = code .. self:_transBody(stmt.body)
  code = code .. self:_transCatch(stmt.catch)
  code = code .. self:_transFinally(stmt.finally)

  --(2) return
  return code
end

--Transform a for each statement.
--
--@return string
function StmtTrans:_transForEach(node)
  local trans = self._.trans
  local code

  --(1) transform
  if node.key then
    code = string.format(
      "for (%s [%s, %s] of Object.entries(%s)) { ",
      node.decl == "var" and "let" or "const",
      node.key,
      node.value,
      trans:_trans(node.iter)
    )
  else
    code = string.format(
      "for (%s %s of %s) ",
      node.decl == "var" and "let" or "const",
      node.value,
      trans:_trans(node.iter)
    )
  end

  if node.catch or node.finally then
    code = code .. "try "
  end

  code = code .. self:_transBody(node.body)
  code = code .. self:_transCatch(node.catch)
  code = code .. self:_transFinally(node.finally)

  if node.key then
    code = code .. " }"
  end

  --(2) return
  return code
end

--Transform a return statement.
--
--@return string
function StmtTrans:_transReturn(node)
  local trans = self._.trans
  local code

  --(1) transform
  code = "return"

  if #node.values == 1 then
    code = code .. " " .. trans:_trans(node.values[1])
  end

  code = code .. ";"

  --(2) return
  return code
end

--Transform a yield statement.
--
--@return string
function StmtTrans:_transYield(node)
  return "yield" .. " " .. self._.trans:_trans(node.value) .. ";"
end

--Transform a struct statement.
--
--@param stmt:StructStmt  The statemento to transform.
--@return string
function StmtTrans:_transStruct(stmt)
  local code
  local trans = self._.trans

  --(1) transform
  --name : base
  code = string.format("const $%s = class %s", stmt.name, stmt.name)

  if stmt.base then
    code = code .. " extends " .. stmt.base
  end

  code = code .. " {\n"

  --constructor
  code = code .. string.format("  constructor(fields) {\n")

  if stmt.base then
    if stmt.base == "dogma.EventEmitter" then
      code = code .. string.format("    super();\n")
    else
      code = code .. string.format("    super(fields);\n")
    end
  end

  code = code .. string.format("    if (fields == null) fields = {};\n")

  for _, fld in ipairs(stmt.fields) do
    --expect()
    if fld.optional then
      if fld.type then
        code = code .. string.format(
          "    if (fields['%s'] != null) expect('%s', fields['%s'], %s);\n",
          fld.name,
          fld.name,
          fld.name,
          fld.type
        )
      end
    else
      code = code .. string.format(
        "    expect('%s', fields['%s'], %s);\n",
        fld.name,
        fld.name,
        fld.type or "null"
      )
    end

    --define
    if fld.kind == "const" then
      code = code .. string.format(
        "    Object.defineProperty(this, '%s%s', {value: coalesce(fields['%s'], %s), writable: false, enumerable: %s});\n",
        self:getVisibPrefix(fld.visib),
        fld.name,
        fld.name,
        fld.value and trans:_trans(fld.value) or "null",
        fld.visib == "pub" and "true" or "false"
      )
    else
      code = code .. string.format(
        "    Object.defineProperty(this, '%s%s', {value: coalesce(fields['%s'], %s), writable: true, enumerable: %s});\n",
        self:getVisibPrefix(fld.visib),
        fld.name,
        fld.name,
        fld.value and trans:_trans(fld.value) or "null",
        fld.visib == "pub" and "true" or "false"
      )
    end
  end

  code = code .. string.format(
    "    if (this.%s__post__ instanceof Function) this.%s__post__();\n",
    self:getVisibPrefix("pvt"),
    self:getVisibPrefix("pvt")
  )
  code = code .. "  }\n"

  --}
  code = code .. "}\n"

  --class proxy
  if stmt.annots:has("abstract") then
    code = code .. string.format(
      [[const %s = new Proxy($%s, { apply(receiver, self, args) { /* istanbul ignore next */ throw("'%s' is abstract."); } });]],
      stmt.name,
      stmt.name,
      stmt.name
    )
  else
    code = code .. string.format(
      "const %s = new Proxy($%s, { apply(receiver, self, args) { return new $%s(...args); } });",
      stmt.name,
      stmt.name,
      stmt.name
    )
  end

  if stmt.visib == "export" then
    code = code .. string.format("\nmodule.exports = exports = %s;", stmt.name)
  elseif stmt.visib == "pub" then
    code = code .. string.format("\nmodule.exports['%s'] = exports['%s'] = %s;", stmt.name, stmt.name, stmt.name)
  end

  --Self
  if stmt.annots:has("Self") then
    code = code .. string.format("\nconst Self = %s;", stmt.name)
  end

  --methods
  for _, meth in ipairs(stmt.methods) do
    code = code .. "\n"

    if meth.subtype == StmtType.FN then
      code = code .. self:_transTypeFn(meth)
    elseif meth.subtype == StmtType.PROC then
      code = code .. self:_transTypeProc(meth)
    end

    if meth.annots:find("post") then
      code = code .. string.format(
        "\n%s.prototype.%s__post__ = %s.prototype.%s%s;",
        meth.itype,
        self:getVisibPrefix("pvt"),
        meth.itype,
        self:getVisibPrefix(meth.visib),
        meth.name
      )
    end
  end

  --(2) return
  return code
end

--Transform a type statement.
--
--@return string
function StmtTrans:_transType(stmt)
  if stmt.body then
    return self:_transTypeWithExplicitConstructor(stmt)
  else
    return self:_transTypeWithImplicitConstructor(stmt)
  end
end

function StmtTrans:_transTypeWithExplicitConstructor(stmt)
  local code

  --(1) transform
  --$class
  code = string.format("const $%s = class %s", stmt.name, stmt.name)
  if stmt.base then
    code = code .. " extends " .. stmt.base
  end

  code = code .. " {\n"

  if stmt.annots:has("todo") then
    code = code .. string.format("  constructor() { todo(); }\n")
  else
    if stmt.annots:has("mixable") then
      code = code .. string.format(
        "  constructor() { this._constructor(...arguments); } _constructor(%s) { ",
        self:_transParams(stmt.params)
      )
    else
      code = code .. string.format("  constructor(%s) { ", self:_transParams(stmt.params))
    end
    code = code .. self:_transParamsCheck(stmt.params)
    code = code .. self:_transParamsTrans(stmt.params)
    code = code .. self:_transParamsUnpacks(stmt.params)
    code = code .. self:_transSuperConstructor(stmt.bargs)
    if stmt.annots:has("component") then
      code = code .. string.format("dogma.autobind(this, $%s);", stmt.name)
    end
    if stmt.mixin then
      code = code .. self:_transMixinConstructor(stmt)
    end
    code = code .. self:_transSelfParams(stmt.params)
    if stmt.catch or stmt.finally then code = code .. " try " end
    code = code .. self:_transBody(stmt.body)
    code = code .. self:_transCatch(stmt.catch)
    code = code .. self:_transFinally(stmt.finally)
    code = code .. "  }\n"
  end

  code = code .. "};"
  if stmt.annots:has("component") then
    code = code .. string.format(
      "Object.defineProperty($%s, '__bind__', {value: [], enumerable: false, writable: false});",
      stmt.name
    )
  end

  if stmt.mixin then
    code = code .. string.format("dogma.mixin($%s, %s);", stmt.name, stmt.mixin)
  end

  code = code .. "\n"

  --class proxy
  if stmt.annots:has("abstract") then
    code = code .. string.format(
      [[const %s = new Proxy($%s, { apply(receiver, self, args) { /* istanbul ignore next */ throw("'%s' is abstract."); } });]],
      stmt.name,
      stmt.name,
      stmt.name
    )
  else
    code = code .. string.format(
      "const %s = new Proxy($%s, { apply(receiver, self, args) { return new $%s(...args); } });",
      stmt.name,
      stmt.name,
      stmt.name
    )
  end

  if stmt.visib == "export" then
    code = code .. string.format("module.exports = exports = %s;", stmt.name)
  end

  --Self
  if stmt.annots:has("Self") then
    code = code .. string.format("\nconst Self = %s;", stmt.name)
  end

  --(2) return
  return code
end

function StmtTrans:_transTypeWithImplicitConstructor(stmt)
  local code

  --(1) transform
  --$class
  code = string.format("const $%s = class %s {\n", stmt.name, stmt.name)
  code = code .. string.format(
    "  constructor($obj) { %s/* istanbul ignore next */if ($obj == null) $obj = {};",
    stmt.annots:has("component") and string.format("dogma.autobind(this, $%s);", stmt.name) or ""
  )
  if stmt.mixin then
    code = code .. self:_transMixinConstructor(stmt)
  end
  for _, p in ipairs(stmt.params) do
    if p.optional then
      if p.type then
        code = code .. string.format(
          [[/* istanbul ignore next */if ($obj["%s"] != null) dogma.paramExpected("%s", $obj["%s"], %s);]],
          p.name,
          p.name,
          p.name,
          p.type
        )
      end
    else  --mandatory
      if p.value == nil then
        code = code .. string.format(
          [[/* istanbul ignore next */ dogma.paramExpected("%s", $obj["%s"], %s);]],
          p.name,
          p.name,
          p.type or "null"
        )
      else
        code = code .. string.format(
          [[/* istanbul ignore next */if ($obj["%s"] != null) dogma.paramExpected("%s", $obj["%s"], %s);]],
          p.name,
          p.name,
          p.name,
          p.type or "null"
        )
      end
    end

    if p.modifier == "." then
      code = code .. string.format(
        [[Object.defineProperty(this, "%s", {value: $obj["%s"]%s, enum: true, writable: %s});]],
        transName(p.name),
        transName(p.name),
        p.value and ("||" .. self._.trans:_trans(p.value)) or "",
        not p.const
      )
    else
      code = code .. string.format(
        [[Object.defineProperty(this, "%s%s", {value: $obj["%s"]%s, writable: %s});]],
        self:getVisibPrefix(p.modifier),
        p.name,
        p.name,
        p.value and ("||" .. self._.trans:_trans(p.value)) or "",
        not p.const
      )
    end
  end
  code = code .. "  }\n"

  code = code .. "};"
  if stmt.annots:has("component") then
    code = code .. string.format(
      "Object.defineProperty($%s, '__bind__', {value: [], enumerable: false, writable: false});",
      stmt.name
    )
  end

  if stmt.mixin then
    code = code .. string.format("dogma.mixin($%s, %s);", stmt.name, stmt.mixin)
  end

  code = code .. "\n"

  --class proxy
  if stmt.annots:has("abstract") then
    code = code .. string.format(
      [[const %s = new Proxy($%s, { apply(receiver, self, args) { /* istanbul ignore next */ throw("'%s' is abstract."); } });]],
      stmt.name,
      stmt.name,
      stmt.name
    )
  else
    code = code .. string.format(
      "const %s = new Proxy($%s, { apply(receiver, self, args) { return new $%s(...args); } });",
      stmt.name,
      stmt.name,
      stmt.name
    )
  end

  if stmt.visib == "export" then
    code = code .. string.format("module.exports = exports = %s;", stmt.name)
  end

  --Self
  if stmt.annots:has("Self") then
    code = code .. string.format("\nconst Self = %s;", stmt.name)
  end

  --(2) return
  return code
end

--Transform a mixin constructor call.
--
--@param stmt type statement.
--@return string
function StmtTrans:_transMixinConstructor(stmt)
  local trans = self._.trans
  local args

  if stmt.margs[1] == "..." then
    args = "...arguments"
  else
    args = ""

    for _, arg in ipairs(stmt.margs) do
      args = args .. (args == "" and "" or ", ") .. trans:_trans(arg)
    end
  end

  return string.format(
    "/* istanbul ignore next */if (%s.prototype._constructor instanceof Function) %s.prototype._constructor.bind(this)(%s);",
    stmt.mixin,
    stmt.mixin,
    args
  )
end

--Transform function parameters.
--
--@return string
function StmtTrans:_transParams(params)
  local trans = self._.trans
  local code

  --(1) transform
  if #params > 0 then
    code = ""

    for i, p in ipairs(params) do
      if i > 1 then
        code = code .. ", "
      end

      if p.modifier == "..." then
        code = code .. "..."
      end

      code = code .. transName(p.name)

      if p.value then
        code = code .. " = " .. trans:_trans(p.value)
      end
    end
  end

  --(2) return
  return code or ""
end

--Return the code for checking the function parameters.
--
--@return string
function StmtTrans:_transParamsCheck(params)
  local function transType(obj)
    local repr

    repr = "{"
    for ix, val in ipairs(obj) do
      local dtype = val.type

      if type(dtype) ~= "string" then
        dtype = string.format("[%s]", table.concat(dtype, ", ")):gsub("list%*", "list")
      end

      repr = repr .. (ix == 1 and "" or ", ") .. string.format(
        "%s: {type: %s, mandatory: %s}",
        val.name,
        dtype,
        val.mandatory
      )
    end
    repr = repr .. "}"

    return repr
  end

  local code

  --(1) transform
  if #params > 0 then
    code = ""

    for _, p in ipairs(params) do
      if not p.optional and p.modifier ~= "..." then  --mandatory parameter with(out) type check
        if type(p.type) == "table" then
          if type(p.type[1]) == "string" then
            code = code .. string.format(
              [[/* istanbul ignore next */ dogma.paramExpected("%s", %s, [%s]);]],
              transName(p.name),
              transName(p.name),
              table.concat(p.type, ", "):gsub("list%*", "list")
            )
          else
            code = code .. string.format(
              [[/* istanbul ignore next */ dogma.paramExpectedToHave("%s", %s, %s, true);]],
              transName(p.name),
              transName(p.name),
              transType(p.type)
            )
          end
        else
          code = code .. string.format(
            [[/* istanbul ignore next */ dogma.paramExpected("%s", %s, %s);]],
            transName(p.name),
            transName(p.name),
            p.type or "null"
          )
        end
      elseif p.type then  --optional parameter with type check
        if type(p.type) == "table" then
          if type(p.type[1]) == "string" then
            code = code .. string.format(
              [[/* istanbul ignore next */ dogma.paramExpectedToBe("%s", %s, [%s]);]],
              transName(p.name),
              transName(p.name),
              table.concat(p.type, ", "):gsub("list%*", "list")
            )
          else
            code = code .. string.format(
              [[/* istanbul ignore next */ dogma.paramExpectedToHave("%s", %s, %s, false);]],
              transName(p.name),
              transName(p.name),
              transType(p.type)
            )
          end
        else
          if p.type ~= "list*" then
            code = code .. string.format(
              [[/* istanbul ignore next */ dogma.paramExpectedToBe("%s", %s, %s);]],
              transName(p.name),
              transName(p.name),
              p.type
            )
          end
        end
      end
    end
  end

  --(2) return
  return code or ""
end

--Return the code for transforming function parameters.
--
--@return string
function StmtTrans:_transParamsTrans(params)
  local code = ""

  --(1) transform
  if type(params) == "table" then
    if #params > 0 then
      for _, p in ipairs(params) do
        if p.type == "list*" then
          code = code .. string.format(
            "%s = list(%s);",
            p.name,
            p.name
          )
        elseif type(p.type) == "table" then
          if #p.type > 0 and p.type[1] == "list*" then  --list
            code = code .. string.format(
              "%s%s = list(%s);",
              p.optional and string.format("if (%s != null) ", p.name) or "",
              p.name,
              p.name
            )
          else  --map
            for _, fld in ipairs(p.type) do
              if type(fld.type) == "table" and fld.type[1] == "list*" then
                code = code .. string.format(
                  "%s%s.%s = list(%s.%s);",
                  fld.mandatory and "" or string.format("if (%s.%s != null) ", p.name, fld.name),
                  p.name,
                  fld.name,
                  p.name,
                  fld.name
                )
              end
            end
          end
        end
      end
    end
  end

  --(2) return
  return code
end

--Return the code for transforming parameter unpacks.
--
--@return string
function StmtTrans:_transParamsUnpacks(params)
  local code = ""

  --(1) transform
  if type(params) == "table" and #params > 0 then
    for _, p in ipairs(params) do
      if p.uVars and #p.uVars > 0 then
        code = code .. string.format(
          "let %s%s%s = %s;",
          p.type == "list" and "[" or "{",
          table.concat(p.uVars, ","),
          p.type == "list" and "]" or "}",
          p.name
        )
      end
    end
  end

  --(2) return
  return code
end

--Return the code for setting . or : attributes from parameters.
--
--@return string
function StmtTrans:_transSelfParams(params)
  local code

  --(1) transform
  if #params > 0 then
    code = ""

    for _, p in ipairs(params) do
      if p.modifier == "." then
        code = code .. string.format(
          [[Object.defineProperty(this, "%s", {value: %s, enum: true, writable: %s});]],
          transName(p.name),
          transName(p.name),
          not p.const
        )
      elseif tablex.find({":", "!"}, p.modifier) then
        code = code .. string.format(
          [[Object.defineProperty(this, "%s%s", {value: %s, writable: %s});]],
          self:getVisibPrefix(p.modifier),
          p.name,
          p.name,
          not p.const
        )
      end
    end
  end

  --(2) return
  return code or ""
end

--Return a call to the super constructor.
--
--@return string
function StmtTrans:_transSuperConstructor(bargs)
  local trans = self._.trans
  local code

  --(1) Transform
  if bargs then
    if #bargs == 0 then
      code = "super();"
    elseif #bargs == 1 and bargs[1] == "..." then
      code = "super(...arguments);"
    else
      code = "super("

      for i, a in ipairs(bargs) do
        code = code .. (i == 1 and "" or ", ") .. trans:_trans(a)
      end

      code =  code .. ");"
    end
  end

  --(2) return
  return code or ""
end

--Transform a proc statement.
--
--@return string
function StmtTrans:_transProc(stmt)
  if stmt.itype then
    return self:_transTypeProc(stmt)
  else
    return self:_transStdProc(stmt)
  end
end

function StmtTrans:_transStdProc(fn)
  local code

  --(1) transform
  if fn.annots:has("todo") then
    code = string.format(
      "%sfunction%s %s() { todo(); }",
      fn.async and "async " or "",
      fn.annots:has("iter") and "*" or "",
      fn.name
    )
  else
    code = string.format(
      "%sfunction%s %s(%s) { ",
      fn.async and "async " or "",
      fn.annots:has("iter") and "*" or "",
      fn.name,
      self:_transParams(fn.params)
    )

    if fn.annots:has("input") then
      code = code .. "print(...arguments);"
    end

    code = code .. self:_transParamsCheck(fn.params)
    code = code .. self:_transParamsTrans(fn.params)
    code = code .. self:_transSelfParams(fn.params)
    code = code .. self:_transParamsUnpacks(fn.params)
    if fn.catch or fn.finally then code = code .. " try " end
    code = code .. self:_transBody(fn.body)
    code = code .. self:_transCatch(fn.catch)
    code = code .. self:_transFinally(fn.finally)
    code = code .. " }"
  end

  code = code .. self:_transVisib(fn.visib, fn.name)

  --(2) alias
  if fn.annots:has("alias") then
    local alias = fn.annots:find("alias").args[1]
    code = code .. string.format(" const %s = %s;", alias, fn.name)
    code = code .. self:_transVisib(fn.visib, alias)
  end

  --(3) return
  return code
end

function StmtTrans:_transTypeProc(stmt)
  return self:_transMethodProc(stmt)
end

function StmtTrans:_transMethodProc(stmt)
  local code

  --(1) transform
  if stmt.annots:has("abstract") then
    if stmt.visib == "pvt" then
      self:_raise("abstract method can't be private (!).", stmt)
    end

    code = string.format(
      "/* istanbul ignore next */ %s.prototype.%s%s = function() { abstract(); };",
      stmt.itype,
      self:getVisibPrefix(stmt.visib),
      stmt.name
    )
  else
    if stmt.annots:has("static") then
      if stmt.annots:has("todo") then
        code = string.format(
          "%s.%s%s = %sfunction() { todo(); };",
          stmt.itype,
          self:getVisibPrefix(stmt.visib),
          stmt.name,
          stmt.async and "async " or ""
        )
      else
        code = string.format(
          "%s.%s%s = %sfunction(%s) { ",
          stmt.itype,
          self:getVisibPrefix(stmt.visib),
          stmt.name,
          stmt.async and "async " or "",
          self:_transParams(stmt.params)
        )
      end
    else
      if stmt.annots:has("todo") then
        code = string.format(
          "%s.prototype.%s%s = function() { todo(); };",
          stmt.itype,
          self:getVisibPrefix(stmt.visib),
          stmt.name
        )
      elseif stmt.annots:has("bind") then
        code = string.format(
          "%s.__bind__.push(['%s%s', %sfunction(%s) { ",
          stmt.itype,
          self:getVisibPrefix(stmt.visib),
          stmt.name,
          stmt.async and "async " or "",
          self:_transParams(stmt.params)
        )
      else
        code = string.format(
          "%s.prototype.%s%s = %sfunction(%s) { ",
          stmt.itype,
          self:getVisibPrefix(stmt.visib),
          stmt.name,
          stmt.async and "async " or "",
          self:_transParams(stmt.params)
        )

        if stmt.annots:has("alias") then
          code = string.format("%s.prototype.%s = ", stmt.itype, stmt.annots:find("alias").args[1]) .. code
        end
      end
    end

    if not stmt.annots:has("todo") then
      if stmt.annots:has("input") then
        code = code .. "print(...arguments);"
      end
      code = code .. self:_transParamsCheck(stmt.params)
      code = code .. self:_transParamsTrans(stmt.params)
      code = code .. self:_transParamsUnpacks(stmt.params)
      code = code .. self:_transSelfParams(stmt.params)

      if stmt.catch or stmt.finally then code = code .. " try " end
      code = code .. self:_transBody(stmt.body)
      code = code .. self:_transCatch(stmt.catch)
      code = code .. self:_transFinally(stmt.finally)

      code = code .. " }"
      if stmt.annots:has("bind") then
        code = code .. "])"
      end
      code = code .. ";"
    end
  end

  --(2) return
  return code
end

--Transform a fn statement.
--
--@return string
function StmtTrans:_transFn(stmt)
  if stmt.itype then
    return self:_transTypeFn(stmt)
  else
    return self:_transStdFn(stmt)
  end
end

function StmtTrans:_transStdFn(fn)
  local code

  --(1) transform
  if fn.annots:has("todo") then
    code = string.format(
      "%sfunction%s %s() { todo(); }",
      fn.async and "async " or "",
      fn.annots:has("iter") and "*" or "",
      fn.name
    )
  else
    code = string.format(
      "%sfunction%s %s(%s) { ",
      fn.async and "async " or "",
      fn.annots:has("iter") and "*" or "",
      fn.name,
      self:_transParams(fn.params)
    )

    if fn.annots:has("input") then
      code = code .. "print(...arguments);"
    end

    code = code .. self:_transReturnVar(fn)
    code = code .. self:_transParamsCheck(fn.params)
    code = code .. self:_transParamsTrans(fn.params)
    code = code .. self:_transSelfParams(fn.params)
    code = code .. self:_transParamsUnpacks(fn.params)
    if fn.catch or fn.finally then code = code .. " try " end
    code = code .. self:_transBody(fn.body)
    code = code .. self:_transCatch(fn.catch)
    code = code .. self:_transFinally(fn.finally)
    if fn.rvar then
      code = code .. string.format(" return %s;", fn.rvar == "self" and "this" or fn.rvar)
    end
    code = code .. " }"
  end

  code = code .. self:_transVisib(fn.visib, fn.name)

  --(2) alias
  if fn.annots:has("alias") then
    local alias = fn.annots:find("alias").args[1]
    code = code .. string.format(" const %s = %s;", alias, fn.name)
    code = code .. self:_transVisib(fn.visib, alias)
  end

  --(3) return
  return code
end

function StmtTrans:_transTypeFn(stmt)
  if stmt.annots:has("prop") then
    return self:_transProp(stmt)
  elseif stmt.annots:has("iter") then
    return self:_transIterMethod(stmt)
  else
    return self:_transMethod(stmt)
  end
end

function StmtTrans:_transIterMethod(stmt)
  local code

  --(1) transform
  if stmt.annots:has("todo") then
    if stmt.name == "iter" then
      code = string.format(
        "%s.prototype[Symbol.iterator] = function*() { todo(); };",
        stmt.itype
      )
    else
      code = string.format(
        '%s.prototype.%s%s = function*() { todo(); };',
        stmt.itype,
        stmt.visib == "pub" and "" or "_",
        stmt.name
      )
    end
  else
    if stmt.name == "iter" then
      code = string.format(
        "%s.prototype[Symbol.iterator] = function*() ",
        stmt.itype
      )
    else
      code = string.format(
        '%s.prototype.%s%s = function*() ',
        stmt.itype,
        stmt.visib == "pub" and "" or "_",
        stmt.name
      )
    end

    code = code .. self:_transReturnVar(stmt)
    if stmt.catch or stmt.finally then code = code .. " try " end
    code = code .. self:_transBody(stmt.body)
    code = code .. self:_transCatch(stmt.catch)
    code = code .. self:_transFinally(stmt.finally)

    code = code .. ";"
  end

  --(2) return
  return code
end

function StmtTrans:_transProp(stmt)
  local code

  --(1) transform
  code = string.format(
    [[Object.defineProperty(%s.prototype, "%s%s", {enum: %s, get: function() { ]],
    stmt.itype,
    stmt.visib == "pub" and "" or "_",
    stmt.name,
    stmt.visib == "pub"
  )

  if stmt.annots:has("abstract") then
    code = code .. "/* istanbul ignore next */abstract();"
  elseif stmt.annots:has("todo") then
    code = code .. "todo();"
  else
    code = code .. self:_transReturnVar(stmt)
    if stmt.catch or stmt.finally then code = code .. " try " end
    code = code .. self:_transBody(stmt.body)
    code = code .. self:_transCatch(stmt.catch)
    code = code .. self:_transFinally(stmt.finally)

    if stmt.rvar then
      code = code .. string.format(" return %s;", stmt.rvar == "self" and "this" or stmt.rvar)
    end
  end

  code = code .. " }});"

  --(2) return
  return code
end

function StmtTrans:_transMethod(stmt)
  local code

  --(1) transform
  if stmt.annots:has("abstract") then
    if stmt.visib == "pvt" then
      self:_raise("abstract method can't be private (!).", stmt)
    end

    code = string.format(
      "/* istanbul ignore next */ %s.prototype.%s%s = function() { abstract(); };",
      stmt.itype,
      self:getVisibPrefix(stmt.visib),
      stmt.name
    )
  else
    if stmt.annots:has("static") then
      if stmt.annots:has("todo") then
        code = string.format(
          "%s.%s%s = %sfunction() { todo(); };",
          stmt.itype,
          self:getVisibPrefix(stmt.visib),
          stmt.name,
          stmt.async and "async " or ""
        )
      else
        code = string.format(
          "%s.%s%s = %sfunction(%s) { ",
          stmt.itype,
          self:getVisibPrefix(stmt.visib),
          stmt.name,
          stmt.async and "async " or "",
          self:_transParams(stmt.params)
        )

        if stmt.annots:has("alias") then
          code = string.format("%s.%s = ", stmt.itype, stmt.annots:find("alias").args[1]) .. code
        end
      end
    else
      if stmt.annots:has("todo") then
        code = string.format(
          "%s.prototype.%s%s = function() { todo(); };",
          stmt.itype,
          self:getVisibPrefix(stmt.visib),
          stmt.name
        )
      elseif stmt.annots:has("bind") then
        code = string.format(
          "%s.__bind__.push(['%s%s', %sfunction(%s) { ",
          stmt.itype,
          self:getVisibPrefix(stmt.visib),
          stmt.name,
          stmt.async and "async " or "",
          self:_transParams(stmt.params)
        )
      else
        code = string.format(
          "%s.prototype.%s%s = %sfunction(%s) { ",
          stmt.itype,
          self:getVisibPrefix(stmt.visib),
          stmt.name,
          stmt.async and "async " or "",
          self:_transParams(stmt.params)
        )

        if stmt.annots:has("alias") then
          code = string.format("%s.prototype.%s = ", stmt.itype, stmt.annots:find("alias").args[1]) .. code
        end
      end
    end

    if not stmt.annots:has("todo") then
      if stmt.annots:has("input") then
        code = code .. "print(...arguments);"
      end
      code = code .. self:_transReturnVar(stmt)
      code = code .. self:_transParamsCheck(stmt.params)
      code = code .. self:_transParamsTrans(stmt.params)
      code = code .. self:_transParamsUnpacks(stmt.params)
      code = code .. self:_transSelfParams(stmt.params)

      if stmt.catch or stmt.finally then code = code .. " try " end
      code = code .. self:_transBody(stmt.body)
      code = code .. self:_transCatch(stmt.catch)
      code = code .. self:_transFinally(stmt.finally)

      if stmt.rvar then
        code = code .. string.format(" return %s;", stmt.rvar == "self" and "this" or stmt.rvar)
      end

      code = code .. " }"
      if stmt.annots:has("bind") then
        code = code .. "])"
      end
      code = code .. ";"
    end
  end

  --(2) return
  return code
end

function StmtTrans:_transReturnVar(fn)
  local code

  --(1) transform
  code = ""

  if fn.rvar then
    if fn.rvar ~= "self" and not fn.params:has(fn.rvar) then
      if fn.rtype == "map" then
        code = string.format("let %s = {};", fn.rvar)
      elseif fn.rtype == "list" then
        code = string.format("let %s = [];", fn.rvar)
      elseif fn.rtype == "set" then
        code = string.format("let %s = set();", fn.rvar)
      elseif fn.rtype == "bool" then
        code = string.format("let %s = false;", fn.rvar)
      elseif fn.rtype == "text" then
        code = string.format('let %s = "";', fn.rvar)
      elseif fn.rtype == "num" then
        code = string.format('let %s = 0;', fn.rvar)
      else
        code = string.format("let %s;", fn.rvar)
      end
    end
  end

  --(2) return
  return code
end

--Transform a pub statement.
--
--@return string
function StmtTrans:_transPub(stmt)
  local code

  --(1) transform
  code = ""

  for _, item in ipairs(stmt.items) do
    if item.type == "pub" then
      code = code .. string.format("export {%s};", item.value)
    else
      code = code .. self:_transUse({modules = {item.value}})
      code = code .. string.format("export {%s};", item.value.name)
    end
  end

  --(2) return
  return code
end

--Transform an export statement.
--
--@return string
function StmtTrans:_transExport(stmt)
  local trans = self._.trans
  return "module.exports = exports = " .. trans:_trans(stmt.exp) .. ";"
end

--Transform a with statement.
--
--@return string
function StmtTrans:_transWith(stmt, covDir)
  local trans = self._.trans
  local code

  --(1) transform
  code = string.format("{const _ = %s;", trans:_trans(stmt.value))

  if covDir ~= nil and covDir.ignore ~= "else" then
    code = code .. trans._.directiveTrans:transform(covDir)
  end

  code = code .. " switch (_) {"

  for _, cls in ipairs(stmt.ifs) do
    code = code .. string.format(
      " case %s: %s; /* istanbul ignore next */break;",
      trans:_trans(cls.cond),
      self:_transBody(cls.body)
    )
  end

  if stmt.els then
    if covDir ~= nil and covDir.ignore == "else" then
      covDir.ignore = "next"
      code = code .. " " .. trans._.directiveTrans:transform(covDir)
    end

    code = code .. string.format(" default: %s", self:_transBody(stmt.els))
  end

  code = code .. "}}"

  --(2) return
  return code
end

--Transform an if statement.
--
--@return string
function StmtTrans:_transIf(stmt, covDir)
  local trans = self._.trans
  local code

  --(1) transform
  if stmt.value then
    local exp = stmt.value.tree.root

    if exp.type == NodeType.NON_TERMINAL and  --workaround: resolve in Go version
       exp.subtype == NonTerminalType.OP and
       exp.arity == "b" and
       (exp.op == ":=" or exp.op == "::=") and
       exp.children[1].type == NodeType.TERMINAL and
       exp.children[1].subtype == TerminalType.LIST then
      stmt.decl = (exp.op == ":=" and "var" or "const")
      code = "{"
      if covDir and covDir.ignore == "next" then
        code = code .. trans._.directiveTrans:transform(covDir)
      end
      code = code .. string.format("%s [", (stmt.decl == "var" and "let" or "const"))
      for i, name in ipairs(exp.children[1].data) do
        code = code .. (i == 1 and "" or ", ") .. name.tree.root.data
      end
      code = code .. "]"
      code = code .. " = " .. trans._.expTrans:_transNode(exp.children[2]) .. "; "
    elseif stmt.decl then
      code = "{"
      if covDir and covDir.ignore == "next" then
        code = code .. trans._.directiveTrans:transform(covDir)
      end
      code = code .. (stmt.decl == "var" and "let " or "const ")
      code = code .. trans:_trans(stmt.value):sub(2, -2) .. "; "
    else
      if covDir and covDir.ignore == "next" then
        code = trans._.directiveTrans:transform(covDir)
      else
        code = ""
      end

      code = code .. trans:_trans(stmt.value) .. "; "
    end
  else
    code = ""
  end

  if covDir ~= nil then
    code = code .. trans._.directiveTrans:transform(covDir)
  end

  code = code .. string.format(
    "if (%s) %s",
    trans:_trans(stmt.cond),
    self:_transBody(stmt.body)
  )

  if stmt.elif then
    for _, cl in ipairs(stmt.elif) do
      code = code .. string.format(
        " else if (%s) %s",
        trans:_trans(cl.cond),
        self:_transBody(cl.body)
      )
    end
  end

  if stmt.el then
    code = code .. " else " .. self:_transBody(stmt.el)
  end

  if stmt.decl then
    code = code .. "}"
  end

    --(2) return
  return code
end
